#!/usr/bin/python3

# data_list.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import os, re, gi
gi.require_version(namespace='Adw', version='1')
from gi.repository import GObject, Gio

class DataList():
    """
    A DataList consists of DataItems.
    Consists means the DataList has no copy of the DataItem, but points to the DataItem.
    The main DataList (main_list) consists of all DataItems in a sorted order (sorted by the user).

    Additional DataLists (plus_list), that may be needed in Widgets, can be added and may
    have automatically calculated and updated parts of the DataItems
    have a changed order of the DataItems,
    or have an added first data that represents a text (line1).

    Advantages:
      * If one DataItem is changed in any DataList, it is automatically changed
        in all other DataLists.
    """

    def __init__(self):
        self.main_list = Gio.ListStore()
        self.plus_lists = []
        self.main_list_last = self.add_plus_list() # needed to find out deleted DI
        self.main_list.connect('items-changed', self.on_data_list_changed)

    def str(self, data_list):
        return "[" + ",".join([str(data) for data in data_list]) + "]"

    def __str__(self):
        return self.str(self.main_list)

    def add_plus_list(self, pick=None, sort=False, item1=None, selfname=None):
        """
        pick  None: to contain all DataItem
              function: return True if the given DataItem is picked of False
        sort  True: sort added DataItem in the same order as main_list
              False: add new DataItem to the end
        item1   None: No additional first DataItem
              DataItem: The given DataItem should represent a string if printed
        """
        plus=Gio.ListStore()
        if item1:
            plus.append(item1)
        self.plus_lists.append((plus, pick, sort))
        if selfname:
            setattr(self, selfname, plus)
        return plus

    def on_item_changed(self, data, field):
        self.add_item_to_plus_list_where_needed(data)

    def add_item_to_plus_list_where_needed(self, item):
        for plus_list, pick, sort in self.plus_lists:
            found, pos = plus_list.find(item)
            if pick:
                if pick(item):
                    if not found:
                        if sort:
                            self.add_item_sorted(item, plus_list)
                        else:
                            plus_list.append(item)
                elif found:
                    plus_list.remove(pos)
            elif not found:
                if sort:
                    self.add_item_sorted(item, plus_list)
                else:
                    plus_list.append(item)

    def del_item_from_plus_list_where_needed(self, item):
        for plus_list, _, _ in self.plus_lists:
            found, pos = plus_list.find(item)
            if found:
                plus_list.remove(pos)


    def add_item_sorted(self, item, plus_list, order_list=None):
        """
        >>> order_list = [1,2,3,4,5,6,7,8]
        >>> plus_list = [2,4,6,8]
        >>> DataList.add_item_sorted(None, 5, plus_list, order_list)
        >>> DataList.str(None, plus_list)
        '[2,4,5,6,8]'

        >>> plus_list = ["first",2,4,6,8]
        >>> DataList.add_item_sorted(None, 5, plus_list, order_list)
        >>> DataList.str(None, plus_list)
        '[first,2,4,5,6,8]'

        """
        if order_list is None:
            order_list=self.main_list

        is_break=False
        plus_i=0
        order_i=0

        if len(plus_list) >=1 and not plus_list[0] in order_list:
            # special item1 used
            plus_i=1

        while plus_i < len(plus_list):
            plus_item = plus_list[plus_i]
            while order_i < len(order_list):
                order_item = order_list[order_i]
                if order_item == item:
                    is_break=True
                    break
                if order_item == plus_item:
                    break
                order_i+=1
            if is_break:
                break
            plus_i+=1
        plus_list.insert(plus_i, item)

    def on_data_list_changed(self, list_store, position, removed, added):
        assert added==1 or removed==1
        if added==1:
            item = self.main_list[position]
            item.connect('notify', self.on_item_changed)
            self.add_item_to_plus_list_where_needed(item)
        elif removed==1:
            item = self.main_list_last[position]
            self.del_item_from_plus_list_where_needed(item)

    def clear(self):
        for _ in range(0, len(self.main_list)):
            # currently only deleting one item at once is allowed by on_data_list_changed
            self.main_list.remove(0)

if __name__ == '__main__':

    import doctest
    doctest.testmod()

    class Data(GObject.Object):
        name = GObject.Property(type=str)
        count = GObject.Property(type=int)

        def __init__(self, name, count=0):
            super().__init__()
            self.name = name
            self.count = count

            self.picklists=[]
            self.unpicklists=[]

        def __str__(self):
            if self.count >=0:
                return f"({self.name},{self.count})"
            # used for DataList.plus_lists to display item1 in a drop down box
            return f"{self.name}"

    def is_picked(item):
        return item.count > 0

    def is_unpicked(item):
        return not is_picked(item)

    data = DataList()
    main_list = data.main_list
    data.add_plus_list(pick=is_picked, selfname="picked1")
    unpicked1 = data.add_plus_list(pick=is_unpicked)
    unpicked2 = data.add_plus_list(pick=is_unpicked, sort=True, item1=Data("select item",-1))
    assert f"{data}" == "[]"
    assert f"{data.str(data.main_list)}" == "[]"
    assert f"{data.str(data.main_list_last)}" == "[]"
    assert f"{data.str(data.picked1)}" == "[]"
    assert f"{data.str(unpicked1)}" == "[]"
    assert f"{data.str(unpicked2)}" == "[select item]"

    for (a,b) in [("one",1),("two",2),("three",3),("four",0)]:
        main_list.append(Data(a,b))
    assert f"{data}" == "[(one,1),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.main_list)}" == "[(one,1),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.main_list_last)}" == "[(one,1),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.picked1)}" == "[(one,1),(two,2),(three,3)]"
    assert f"{data.str(unpicked1)}" == "[(four,0)]"
    assert f"{data.str(unpicked2)}" == "[select item,(four,0)]"

    main_list[0].count=0
    assert f"{data}" == "[(one,0),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.main_list)}" == "[(one,0),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.main_list_last)}" == "[(one,0),(two,2),(three,3),(four,0)]"
    assert f"{data.str(data.picked1)}" == "[(two,2),(three,3)]"
    assert f"{data.str(unpicked1)}" == "[(four,0),(one,0)]"
    assert f"{data.str(unpicked2)}" == "[select item,(one,0),(four,0)]"

    main_list.remove(2)
    assert f"{data}" == "[(one,0),(two,2),(four,0)]"
    assert f"{data.str(data.main_list)}" == "[(one,0),(two,2),(four,0)]"
    assert f"{data.str(data.main_list_last)}" == "[(one,0),(two,2),(four,0)]"
    assert f"{data.str(data.picked1)}" == "[(two,2)]"
    assert f"{data.str(unpicked1)}" == "[(four,0),(one,0)]"
    assert f"{data.str(unpicked2)}" == "[select item,(one,0),(four,0)]"

    print("all asserts have been ok")



#!/usr/bin/python3

# cshbx.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import gi, os, sys, re
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir1)
from cashbox.read_appargs import read_appargs, appargs
from cashbox.locale_utils import _, f

class Cshbx():
    """
    a text line with an article needs an article name and and a price
    optionally it can contain a count
    """

    def __init__(self):
        re_price = r'(\$\s*|\€\s*|)(\d+)(,|\.)(\d\d)(\s*\$|\s*€|\s*Dollar|\s*Euro|)'
        re_name = r'(\s*)(.*\S)(\s+)'
        re_count = r'(?:(\s+)(\d+)){0,1}(\s*)$'
        self.re_sale = re.compile(r'^'+re_name+re_price+re_count+r'$')

        self.re_price = re.compile(re_price)
        self.re_name = re.compile(r'^' + re_name + r'$')
        self.re_count = re.compile(r'^' + re_count + r'$')
        self.re_some=re.compile(r'^(\s*)(.*\S)(\s*)$')

    def get_cent_price(self, line):
        res=0
        match = self.re_sale.match(line)
        if match:
            groups=match.groups()
            if groups[9]:
                res+=(int(groups[4])*appargs.cents + int(groups[6])) * int(groups[9])
        return res

if __name__ == '__main__':
    read_appargs("", sys.argv)
    cshbx = Cshbx()

    assert cshbx.re_price.match("1,20").groups() == ('', '1', ',', '20', '')
    assert cshbx.re_price.match("1.20 Dollar").groups() == ('', '1', '.', '20', ' Dollar')
    assert list(cshbx.re_price.finditer("Bier 0,25 Liter 1,10€"))[-1].groups() == ('', '1', ',', '10', '€')
    assert list(cshbx.re_price.finditer("0.25 Dollar"))[-1].groups() == ('', '0', '.', '25', ' Dollar')
    assert list(cshbx.re_price.finditer("$0.25 Dollar"))[-1].groups() == ('$', '0', '.', '25', ' Dollar')

    assert cshbx.get_cent_price("garbage") == 0
    assert cshbx.get_cent_price("a 1,20 5") == 600

    assert cshbx.re_price.match("12,34").groups() == ('', '12', ',', '34', '')
    assert cshbx.re_price.match("12,34 €").groups() == ('', '12', ',', '34', ' €')
    assert cshbx.re_price.match("12.34$").groups() == ('', '12', '.', '34', '$')
    assert cshbx.re_price.match("12.3") is  None

    print("all asserts have been ok")



#!/usr/bin/python3

# utils.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import gi, os, sys
gi.require_version(namespace='Adw', version='1')
from gi.repository import Gio

def print_widget(w, level=1):
    """ print widgets

    >>> box=Gtk.Box()
    >>> label=Gtk.Label(label="mylabel")
    >>> box.append(label)
    >>> print_widget(box,1)
    - <GtkBox>
    -- <GtkLabel: mylabel>

    """

    if w is None:
        txt=w.get_name()
        if txt == "GtkLabel":
            txt=f"{txt}: {w.get_label()}"
        print("{'-'*level} <{txt}>")
        level=level+1
        w=w.get_first_child()
        while w:
            print_widget(w, level)
            w = w.get_next_sibling()

#def reduce_window_size(window, x=720, y=1440):
def reduce_window_size(window, x=350, y=600):
    window.props.default_width = x
    window.props.default_height = y

if __name__ == '__main__':
    import doctest
    doctest.testmod()

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def err(msg, retcode=1):
    eprint(f"Error: {msg}")
    sys.exit(retcode)

def create_action(window_or_application, name, function):
    action = Gio.SimpleAction.new(name, None)
    action.connect("activate", function)
    window_or_application.add_action(action)



#!/usr/bin/python3

# pick_widget.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import gi, os, sys
gi.require_version(namespace='Adw', version='1')
from gi.repository import Gtk, GObject
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir1)
from cashbox.utils import print_widget, reduce_window_size
from cashbox.article import Article
from cashbox.data_list import DataList
from cashbox.locale_utils import _, f


@Gtk.Template(filename='%s/pick_widget_row.ui' % dir1)
class PickRow(Gtk.Box):
    __gtype_name__ = 'PickRow'
    action_row = Gtk.Template.Child()
    spin_button = Gtk.Template.Child()

@Gtk.Template(filename='%s/pick_widget.ui' % dir1)
class PickWidget(Gtk.Box):
    __gtype_name__ = 'PickWidget'
    list_view = Gtk.Template.Child()

    def __init__(self, sale, **kwargs):
        super().__init__(**kwargs)

        # Selection
        ss = Gtk.NoSelection()
        self.list_view.set_model(ss)

        # sale
        self.sale = sale
        ss.set_model(self.sale.picked)

        # factory
        factory = Gtk.SignalListItemFactory()
        self.list_view.set_factory(factory)
        factory.connect("setup", self.factory_setup)
        factory.connect("bind", self.factory_bind)

    def factory_setup(self, fact, item):
        pick_row = PickRow()
        item.set_child(pick_row)

    def factory_bind(self, fact, item):
        article = item.get_item()
        pick_row = item.get_child()

        # make sure count is only editable with + and - 
        # to avoid opening virtual keyboard on librem5
        c = pick_row.spin_button.get_first_child()
        c.set_sensitive(False)

        article.bind_property("name",
           pick_row.action_row, "title",
           GObject.BindingFlags.SYNC_CREATE)

        article.bind_property("count",
           pick_row.spin_button, "value",
           GObject.BindingFlags.SYNC_CREATE |
           GObject.BindingFlags.BIDIRECTIONAL,
           transform_to=self.transform_count_to_spin_button,
           transform_from=self.transform_spin_button_to_count)

    def transform_count_to_spin_button(self, binding, val):
        print(f"transform_count_to_spin_button val=<{val}> type=<{type(val)}>")
        return val

    def transform_spin_button_to_count(self, binding, val):
        print(f"transform_spin_button_to_count val=<{val}> type=<{type(val)}>")
        return int(val)

    def output(self, spin_row, item):
        article = item.get_item()
        # get count from widget (article.count may not be updated in article.count)
        adjustment=spin_row.get_adjustment()
        count=int(adjustment.get_value())
        article.count=count
        #print(f"output: name=<{article.name}> count=<{count}>")
        if count > 0:
            # write price to subtitle
            spin_row.set_subtitle("%s" % (article.price))

if __name__ == '__main__':
    from cashbox.app import App, MinWindow
    from cashbox.article import Article, Sale
    from cashbox.read_appargs import read_appargs, appargs
    from cashbox.read_css import read_css

    class MyMainWindow(MinWindow):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            if appargs.test_small_display:
                reduce_window_size(self)

            self.sale = Sale()

            article = [("Banana",1.10,1), ("Apple",2.0,2), ("Strawberry",2.50,3),
                       ("very long one two three four five six seven eight",9.99,1),
                       ("Pear",3.35,0), ("Watermelon",1.0,5), ("Blueberry",2.0,6)]
            for f in article:
                self.sale.main_list.append(Article( f[0], f[1], f[2] ))

            self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            self.set_content(self.box)

            list_sale_button = Gtk.Button(label="list")
            self.box.append(list_sale_button)
            list_sale_button.connect('clicked', self.list_sale_clicked)

            self.pick_widget = PickWidget(self.sale)
            self.box.append(self.pick_widget)

            quit_button = Gtk.Button(label="quit")
            self.box.append(quit_button)
            quit_button.connect('clicked', self.quit_clicked)

        def list_sale_clicked(self, _list_button):
            print(f"list_sale_clicked: sale=<{self.sale}>")

        def quit_clicked(self, list_button):
            self.close()

    class MyApp(App):

        def on_activate(self, app):
            self.win = MyMainWindow(application=app)
            self.win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# dialog_widget.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import sys, os, gi
gi.require_version('Adw', '1')
gi.require_version('Gtk', '4.0')
from gi.repository import Adw, Gtk
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.locale_utils import _, f

@Gtk.Template(filename=f'{dir1}/dialog_widget.ui')
class DialogWidget(Gtk.Box):
    __gtype_name__ = 'DialogWidget'
    help_widget_dialog = Gtk.Template.Child()
    help_widget_label = Gtk.Template.Child()
    about_widget_dialog = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.abbreviations()

    def abbreviations(self):
        """
        this method should be overwritten, if DialogWidged class
        is used outside of cashbox.
        """
        self.a = self.data(_("article"))
        self.as_ = self.data(_("articles"))
        self.As = self.data(_("Articles"))
        self.C = self.widget(_("Cashbox"))
        self.c = self.data(_("count"))
        self.cs = self.data(_("counts"))
        self.n = self.data(_("name"))
        self.ns = self.data(_("names"))
        self.p = self.data(_("price"))
        self.ps = self.data(_("prices"))
        self.Pl = self.widget(_("Pricelist"))
        self.R = self.widget(_("Receipt"))
        self.S = self.widget(_("Sale"))

    def widget(self,txt):
        return f"<b><span foreground='blue'>{txt}</span></b>"

    def data(self,txt):
        return f"<b><span foreground='green'>{txt}</span></b>"

    def help_dialog(self, win, title, txt):
        dialog=self.help_widget_dialog
        label=self.help_widget_label
        dialog.set_title(title)
        label.set_label(txt)
        dialog.present(win)

    def about_dialog(self, win, txt, version="0.0"):
        about=self.about_widget_dialog
        about.set_comments(txt)
        about.set_developers(["Bernd Schumacher"])
        about.set_version(version)
        about.present(win)


if __name__ == '__main__':

    import sys
    dir1 = os.path.dirname(os.path.realpath(__file__))
    dir2 = os.path.dirname(dir1)
    sys.path.append(dir2)

    a1=_("hallo")

    class DialogWindow(Adw.ApplicationWindow):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            self.props.default_width = 350
            self.props.default_height = 600

            box = Gtk.Box()
            box.set_orientation(Gtk.Orientation.VERTICAL)
            self.set_content(box)

            button = Gtk.Button(label=_("test_help_dialog"))
            box.append(button)
            button.connect('clicked', self.on_test_help_dialog)

            button = Gtk.Button(label=_("test_about"))
            box.append(button)
            button.connect('clicked', self.on_test_about)

        def on_test_about(self, button):
            d = DialogWidget()
            win=self
            d.about_dialog(win, f(_("""\
one two three {d.As}""")))

        def on_test_help_dialog(self, button):
            d = DialogWidget()
            win=self
            d.help_dialog(win, _("test_title"), f(_("""\
one two three
{d.As} can be listed with {d.n} and {d.p} in {d.Pl}.""")))

    class MyApp(Adw.Application):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.connect('activate', self.on_activate)

        def on_activate(self, app):
            self.win = DialogWindow(application=app)
            self.win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# file_widget.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import os, sys, re, gi
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw, Gtk
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.utils import eprint
from cashbox.article import Article, Sale
from cashbox.drop_down_widget import DropDownWidget
from cashbox.read_css import read_css
from cashbox.read_appargs import read_appargs, appargs

@Gtk.Template(filename='%s/file_widget.ui' % dir1)
class FileWidget(Gtk.Box):
    __gtype_name__ = 'FileWidget'
    file_dialog = Gtk.Template.Child()

    def __init__(self, sale, **kwargs):
        super().__init__(**kwargs)
        self.sale=sale

    @Gtk.Template.Callback()
    def on_load(self, button):
        self.file_dialog.open(None, None, self.load_callback)

    def load_callback(self, gtk_file_dialog, g_task):
        g_local_file=gtk_file_dialog.open_finish(g_task)
        with open(g_local_file.get_path(), 'r') as file:
            data = file.read()

    @Gtk.Template.Callback()
    def on_save(self, button):
        self.file_dialog.save(None, None, self.save_callback)

    def save_callback(self, gtk_file_dialog, g_task):
        g_local_file=gtk_file_dialog.save_finish(g_task)
        with open(g_local_file.get_path(), 'w') as file:
            file.write(self.sale.text())

if __name__ == '__main__':
    import sys, doctest
    from app import App, MinWindow

    class FileWindow(MinWindow):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            sale = Sale()

            article = [("Banana", 1.10, 1), ("Apple", 2.0, 0)]
            for f in article:
                sale.main_list.append(Article( f[0], f[1], f[2] ))

            file_widget = FileWidget(sale)
            self.set_content(file_widget)

    class MyApp(App):

        def on_activate(self, app):
            self.win = FileWindow(application=app)
            self.win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# app.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import sys, gi, os
gi.require_version('Gtk', '4.0')
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw, Gtk, Gio, GLib
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.utils import reduce_window_size, create_action
from cashbox.read_appargs import read_appargs, appargs
from cashbox.read_css import read_css
from cashbox.dialog_widget import DialogWidget
from cashbox.locale_utils import _, f

class MinWindow(Adw.ApplicationWindow):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if hasattr(appargs, "t1") and appargs.t1:
            reduce_window_size(self)

class App(Adw.Application):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, application_id=appargs.application_id,
                         flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE |
                         Gio.ApplicationFlags.HANDLES_OPEN, **kwargs)

        application = self
        create_action(application, "about", self.about_action)

        self.connect('activate', self.on_activate)

        self.add_main_option("t1", 0, GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "test_small_display", None)

        self.add_main_option("currency", ord("c"), GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, "Euro or Dollar", None)

        self.add_main_option("css-path", 0, GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, "Path to css file", None)

        # -v is alredy used by doctests
        self.add_main_option("version", 0, GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "say version", None)

        self.set_option_context_parameter_string("FILE")
        self.set_option_context_summary("price list FILE")
        self.set_option_context_description("the price list FILE should contain lines " +
                                            "with a article name followed by a price")

    def do_command_line(self, command_line):
        opts = command_line.get_options_dict().end().unpack()
        args = command_line.get_arguments()

        read_appargs(opts, args[1:])
        read_css()

        if hasattr(appargs, "version") and appargs.version:
            print(appargs.application_version)
            sys.exit(0)

        self.activate()
        return 0

    def about_action(self, action, param):
        d = DialogWidget()
        win=self.get_active_window()
        d.about_dialog(win, f(_("""\
{d.C} memorises the cost of {d.as_} and calculates the total price and \
change. It is intended for small clubs on a celebration, where members are \
not experienced in memorizing prices and doing mental arithmetic.

{d.C} consists of the 3 windows {d.Pl}, {d.S} and {d.R}, that can be switched \
as needed. {d.Pl} can add, delete or modify {d.as_}, {d.S} picks {d.as_} to \
sell and {d.R} shows the total price of the choosen {d.as_}.""")),
            version=appargs.application_version)

if __name__ == '__main__':
    gi.require_version('Gtk', '4.0')
    from gi.repository import Gtk, Adw

    class MyMainWindow(MinWindow):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            box= Gtk.Box()
            self.set_content(box)

            hamburger = Gtk.MenuButton()
            hamburger.set_icon_name("open-menu-symbolic")
            box.append(hamburger)

            popover = Gtk.PopoverMenu()
            hamburger.set_popover(popover)

            menu = Gio.Menu.new()
            menu.append("Show About Dialog", "app.about")
            popover.set_menu_model(menu)

            button = Gtk.Button(label="click me")
            box.append(button)
            button.connect('clicked', self.clicked)

            print(f"main appargs.digits=<{appargs.digits}>")

            for i in appargs.keys():
                print(f"<{i}>=<{appargs[i]}>")

        def clicked(self, button):
            print("clicked")

    class MyApp(App):

        def on_activate(self, app):
            win = MyMainWindow(application=app)
            win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# cashbox.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import sys, gi, os
gi.require_version('Gtk', '4.0')
gi.require_version(namespace='Adw', version='1')
from gi.repository import Gtk
sys.path.append("/usr/share/cashbox/python3")
from cashbox.pick_widget import PickWidget
from cashbox.article import Article, Sale
from cashbox.sale_widget import SaleWidget
from cashbox.view_switch_window import ViewSwitchWindow
from cashbox.receipt_widget import ReceiptWidget
from cashbox.read_css import read_css
from cashbox.app import App, MinWindow
from cashbox.file_widget import FileWidget
from cashbox.pricelist_widget import PricelistWidget
from cashbox.dialog_widget import DialogWidget
from cashbox.read_appargs import read_appargs, appargs
from cashbox.locale_utils import _, f

class MyApp(App):

    def on_activate(self, app):
        win = ViewSwitchWindow(application=app)

        # sale
        self.last_child_name=None
        self.sale = Sale()

        # PricelistWidget
        box=Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=12,
                    margin_end=12, margin_bottom=12,margin_start=12,spacing=12)
        self.pricelist_widget = PricelistWidget(self.sale, win=win)
        box.append(self.pricelist_widget)
        win.stack.add_titled(box, "Pricelist", _("Pricelist"))

        if appargs.moreargs:
            self.pricelist_widget.read_files(appargs.moreargs)

        # SaleWidget
        box=Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=12,
                    margin_end=12, margin_bottom=12,margin_start=12,spacing=12)
        win.stack.add_titled(box, "Sale", _("Sale"))
        sale_widget = SaleWidget(self.sale, win=win)
        box.append(sale_widget)

        # ReceiptWidget
        box=Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=12,
                    margin_end=12, margin_bottom=12,margin_start=12,spacing=12)
        win.stack.add_titled(box, "Receipt", _("Receipt"))
        receipt_widget = ReceiptWidget(self.sale, win=win)
        box.append(receipt_widget)

        win.present()

app = MyApp()
app.run(sys.argv)



#!/usr/bin/python3

# article.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import gi, os, sys, re
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw, Gtk, Gio, GLib, GObject
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.read_appargs import read_appargs, appargs
from cashbox.data_list import DataList
from cashbox.locale_utils import _, f

def cent2str(cents):
    return f"{cents//appargs.cents}{appargs.separator}{cents%appargs.cents:0{appargs.digits}.0f}"

def str2cent(txt, allow_sloppy=False):
    ret=0
    if allow_sloppy:
        txt=re.sub(r"[.,]", appargs.separator, txt)
    try:
        price=float(txt)
    except:
        return ret
    number_of_digits=str(txt)[::-1].find(appargs.separator)
    if number_of_digits <= appargs.digits:
        ret = int(price * appargs.cents)
    return ret

class Article(GObject.Object):
    name = GObject.Property(type=str)
    price = GObject.Property(type=int)
    count = GObject.Property(type=int)

    def __init__(self, name, price, count=0):
        super().__init__()
        self.name = name
        self.price = price
        self.count = count

    def __str__(self):
        return f"({self.name},{cent2str(self.price)},{self.count})"

    def text(self):
        if self.count:
            ret=f"{self.name} {cent2str(self.price)} {self.count}"
        else:
            ret=f"{self.name} {cent2str(self.price)}"
        return ret

class Comment(GObject.Object):
    comment = GObject.Property(type=str)

    def __str__(self):
        return f"{self.comment}"

    def __init__(self, comment):
        super().__init__()
        self.comment = comment

    def text(self):
        return str(self)

class DropDownHead(GObject.Object):
    drop_down_head = GObject.Property(type=str)

    def __str__(self):
        return f"DropDownHead=<{self.drop_down_head}>"

    def __init__(self, drop_down_head):
        super().__init__()
        self.drop_down_head = drop_down_head

    def text(self):
        return str(self)

class Sale(DataList):

    def __init__(self):
        super().__init__()

        # self.main_list
        self.add_plus_list(pick = self.is_picked, selfname="picked") # self.picked
        self.add_plus_list(pick = self.is_unpicked, selfname="unpicked") # self.unpicked
        self.add_plus_list(pick = self.is_unpicked, sort=True,
                           item1=DropDownHead(_("select article")),
                           selfname="drop_unpicked") # self.drop_unpicked

    def is_picked(self, item):
        ret = False
        if type(item) is Article:
            ret = item.count > 0
        return ret

    def is_unpicked(self, item):
        ret = False
        if type(item) is Article:
            ret = item.count == 0
        return ret

    def text(self):
        """
        This will make Sale ediable as text.
        All relevant information is included.
        """
        return "\n".join([data.text() for data in self.main_list])

    def get_article(self, name, picked=None):
        """
        get article with given name or None.
        picked: 
          None: no matter if picked
          False: not picked
          True: picked
        """
        found=None
        for data in self.main_list:
            if isinstance(data, Article):
                if data.name==name:
                    if picked  is True:
                        if data.count>0:
                            found=data
                    elif picked is False:
                        if data.count==0:
                            found=data
                    elif picked is None:
                        found=data
                    break
        return found

    def count_zero(self):
        """
        e.g. start with a new customer, without picked articles
        """
        for article in self.main_list:
            article.count=0

if __name__ == '__main__':
    read_appargs()

    sale = Sale()

    for d in [Comment("one two three"), Article("Banana",110,1),
              Article("Apple",200,2), Article("Strawberry",250,3),
              Article("Pear",335,4), Article("Watermelon",100,5),
              Comment("four,five six"), Article("Blueberry",200,6)]:
        print(f"adding <{d}>")
        sale.main_list.append( d )
    assert f"{sale}" == "[one two three,(Banana,1.10,1),(Apple,2.00,2)," + \
                        "(Strawberry,2.50,3),(Pear,3.35,4)," + \
                        "(Watermelon,1.00,5),four,five six,(Blueberry,2.00,6)]"
    assert f"{sale.text()}" == """one two three
Banana 1.10 1
Apple 2.00 2
Strawberry 2.50 3
Pear 3.35 4
Watermelon 1.00 5
four,five six
Blueberry 2.00 6"""

    # next 2 lines will normally be done by a widget
    a=sale.get_article("Apple", picked=True)
    a.count=0
    assert f"{sale.get_article('Apple', picked=True)}" == "None"
    assert f"{sale.get_article('Apple', picked=False)}" == "(Apple,2.00,0)"

    a.count=3
    assert f"{sale.get_article('Apple', picked=False)}" == "None"
    assert f"{sale.get_article('Apple', picked=True)}" == "(Apple,2.00,3)"
    assert f"{sale.get_article('Apple')}" == "(Apple,2.00,3)"

    assert a.name == "Apple"
    assert a.price == 200
    assert a.count == 3

    print("all asserts have been ok")



#!/usr/bin/python3

# read_css.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import sys, os

import gi
gi.require_version(namespace='Adw', version='1')
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw, Gtk, Gdk
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from cashbox.read_appargs import read_appargs, appargs

def read_css():

    css_provider = Gtk.CssProvider()
    css_provider.load_from_path(appargs.csspath)

    styleContext = Gtk.StyleContext
    display=Gdk.Display.get_default()
    styleContext.add_provider_for_display(display, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

if __name__ == '__main__':

    # test button
    #   should change color of label to red - defined in cashbox.css as "#invalid"

    gi.require_version(namespace='Adw', version='1')
    from gi.repository import Adw, Gtk
    from app import App

    class CssWindow(Adw.Window):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            box = Gtk.Box()
            self.set_content(box)

            button = Gtk.Button(label="button")
            box.append(button)

            label = Gtk.Label(label="label")
            box.append(label)

            button.connect('clicked', self.clicked, label)

        def clicked(self, _button, label):
            if label.get_name()=="invalid":
                label.set_name("normal")
            else:
                label.set_name("invalid")

    class MyApp(App):

        def on_activate(self, app):
            self.win = CssWindow(application=app)
            self.win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# pricelist_widget.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import os, sys, gi
gi.require_version(namespace='Adw', version='1')
from gi.repository import Gtk, Gio
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.utils import eprint, print_widget, create_action
from cashbox.article import Article, Comment, Sale, str2cent
from cashbox.drop_down_widget import DropDownWidget
from cashbox.read_css import read_css
from cashbox.read_appargs import read_appargs, appargs
from cashbox.cshbx import Cshbx
from cashbox.dialog_widget import DialogWidget
from cashbox.locale_utils import _, f

@Gtk.Template(filename='%s/pricelist_widget.ui' % dir1)
class PricelistWidget(Gtk.Box):
    __gtype_name__ = 'PricelistWidget'
    textview = Gtk.Template.Child()
    error = Gtk.Template.Child()
    file_dialog = Gtk.Template.Child() # Gtk.FileDialog
    cashbox_file_filter = Gtk.Template.Child()
    red_tag = Gtk.Template.Child()
    green_tag = Gtk.Template.Child()
    orange_tag = Gtk.Template.Child()

    @Gtk.Template.Callback()
    def on_unmap_all(self, widget):
        self.check_buffer()

    @Gtk.Template.Callback()
    def on_map_all(self, widget):
        # write sale-variable to widget-buffer
        self.buffer.set_text(self.sale.text())

    def __init__(self, sale, win, **kwargs):
        super().__init__(**kwargs)

        self.win=win
        self.sale=sale
        self.buffer=self.textview.get_buffer()

        self.buffer.set_enable_undo(True)

        self.table=self.buffer.get_tag_table()
        self.cshbx=Cshbx()

        create_action(win, "undo", self.on_undo_action)
        create_action(win, "redo", self.on_redo_action)
        create_action(win, "save", self.on_save_dialog)
        create_action(win, "load", self.on_load_dialog)
        create_action(win, "append", self.on_append_dialog)
        create_action(win, "help_workflow", self.on_help_workflow)
        create_action(win, "help_syntax", self.on_help_syntax)

        self.table.add(self.red_tag)
        self.table.add(self.green_tag)
        self.table.add(self.orange_tag)

        self.buffer.connect('changed', self.on_buffer_changed)
        self.check_buffer()
        self.buffer.connect('notify::cursor-position', self.on_cursor)

        self.cursor_seen_in_line=(-1,-1)

        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(self.cashbox_file_filter)

        self.file_dialog.set_filters(filters)
        self.file_dialog.set_initial_folder(Gio.File.new_for_path(appargs.user_app_dir))


    def on_undo_action(self, action, param):
        self.buffer.undo()

    def on_redo_action(self, action, param):
        self.buffer.redo()

    def on_cursor(self, buffer, pos):
        pos=buffer.get_property("cursor-position")
        (a,b)=self.cursor_seen_in_line
        if pos<a or pos>b:
            self.check_buffer()

    def check_buffer_line(self, line):
        # n=article name, p=article price, c=article count
        (n, p, c, n_err, p_err, c_err) = (None, None, None, None, None, None)
        # get the last match of price in a line
        p_iter = list( self.cshbx.re_price.finditer(line) )
        if len(p_iter):
            p=list(p_iter)[-1]
            a_line=line[:p.start(0)]
            n = self.cshbx.re_name.match(a_line)
            if not n:
                n_err = self.cshbx.re_some.match(a_line)

            c_line=line[p.end(0):]
            c = self.cshbx.re_count.match(c_line)
            if not c:
                c_err = self.cshbx.re_some.match(c_line)
        else:
            p_err = self.cshbx.re_some.match(line)
        return (n, p, c, n_err, p_err, c_err)


    def color_text(self, buffer, line_start, n, p, c, n_err, p_err, c_err,
                   double_name=None, name_to_long=None):
        """
        Checks the line in <buffer> starting at <line_start>, that represent an article.
        Uses the regex results <n> (name), <p> (price) and <c> (optional count) and
        the corresponding errors if the regex failed.
        Returns a list of hints to fix the errors.

        green: ok
        orange: the marked text could be used if another part would be fixed
        red: the text itself must be fixed
        """
        def apply_tag(start,end,tag):
            iter_start = buffer.get_iter_at_offset(line_start + start)
            iter_end = buffer.get_iter_at_offset(line_start + end)
            buffer.apply_tag(tag, iter_start, iter_end)

        error=[]
        if p and n and c and not double_name and not name_to_long:
            apply_tag(n.start(2), n.end(2), self.green_tag)
            apply_tag(p.start(1), p.end(5), self.green_tag)
            apply_tag(p.end(5)+c.start(2), p.end(5)+c.end(2), self.green_tag)
        else:
            if p:
                apply_tag(p.start(0), p.end(1), self.orange_tag)
                if n:
                    if name_to_long:
                        error+=[f(_("name {name_to_long} chars to long"))]
                        apply_tag(n.start(2), n.end(2), self.red_tag)
                    elif double_name:
                        error+=[_("name already exists")]
                        apply_tag(n.start(2), n.end(2), self.orange_tag)
                    else:
                        apply_tag(n.start(2), n.end(2), self.orange_tag)
                elif n_err:
                    apply_tag(n_err.start(2), n_err.end(2), self.red_tag)
                    error+=[_("no space after article name")]
                else:
                    error+=[_("no article name")]
                if c:
                    apply_tag(c.start(2)+p.end(4), c.end(2)+p.end(4), self.orange_tag)
                elif c_err:
                    apply_tag(c_err.start(2)+p.end(5), c_err.end(2)+p.end(5), self.red_tag)
                    error+=[_("bad optional count")]
            elif p_err:
                apply_tag(p_err.start(0), p_err.end(0), self.red_tag)
                error+=[_("no price found")]

        return error

    def on_buffer_changed(self, buffer):
        self.check_buffer()

    def check_buffer(self):
        """
        check text buffer and return the current sale represented
        by the the buffer.

        if buffer2sale == True:
            write widget-buffer to sale
        """

        def get_text(start,end):
            return line[start:end]

        self.sale.clear()

        # clear color tags
        (iter1,iter2)=self.buffer.get_bounds()
        self.buffer.remove_all_tags(iter1, iter2)
        text=self.buffer.get_text(iter1, iter2 , False)

        hint=""
        if text == "":
            hint=_("please add lines with article and price")
        else:
            names=[]
            line_start=0 # start of current line in text
            for (line_nr, line) in enumerate(text.splitlines()):

                cursor=self.buffer.get_property("cursor-position")

                if line_start <= cursor <= line_start+len(line):
                    self.cursor_seen_in_line=(line_start,line_start+len(line))

                if line.lstrip()[:1] == "#":
                    # comment starting with "#"
                    self.sale.main_list.append( Comment(line) )
                else:
                    # n=article name, p=article price, c=article count
                    (n, p, c, n_err, p_err, c_err) = self.check_buffer_line(line)

                    double_name=None
                    name_to_long=None
                    if n:
                        name=get_text(n.start(2),n.end(2))
                        if name in names:
                            double_name=name
                        name_to_long = len(name) - appargs.max_name_len
                        if name_to_long <= 0:
                            name_to_long=None
                    error = self.color_text(self.buffer, line_start, n, p, c, n_err, p_err, c_err,
                                            double_name=double_name, name_to_long=name_to_long)

                    if p and n and c and not double_name and not name_to_long:
                        names.append(name)
                        article=Article(name, str2cent(get_text(p.start(2), p.end(2)) +
                                        "." + get_text(p.start(4), p.end(4))),
                                        int("0" + get_text(p.end(5) + c.start(2)
                                        ,p.end(5) + c.end(2))))
                        #eprint(f"article=<{article}>")
                        self.sale.main_list.append(article)
                    else:
                        # error: give error message, but handled like a comment
                        self.sale.main_list.append( Comment(line) )

                    if line_start <= cursor <= line_start+len(line) and error:
                        hint = f"line {line_nr+1}: "+", ".join(error)

                line_start+=len(line)+1

        self.error.set_label(hint)


    def on_save_dialog(self, action, param):
        self.file_dialog.save(None, None, self.on_save_dialog_finish)

    def on_save_dialog_finish(self, dialog, task):
        file=dialog.save_finish(task)

        (iter1,iter2)=self.buffer.get_bounds()
        text=self.buffer.get_text(iter1, iter2 , False)

        path=file.get_path()
        if not path.endswith(appargs.conf_suffix):
            path=path+appargs.conf_suffix

        with open(path, 'w') as p:
            p.write(text)

    def on_load_dialog(self, action, param):
        self.file_dialog.open(None, None, self.on_load_dialog_finish)

    def on_load_dialog_finish(self, dialog, task):
        file=dialog.open_finish(task)
        (iter1,iter2)=self.buffer.get_bounds()
        with open(file.get_path(), 'r') as p:
            data = p.read()
            self.buffer.delete(iter1, iter2)
            self.buffer.insert(iter1, data)

    def on_append_dialog(self, action, param):
        self.file_dialog.open(None, None, self.on_append_dialog_finish)

    def on_help_workflow(self, action, _param):
        d = DialogWidget()
        win=self
        d.help_dialog(win, _("pricelist workflow"), f(_("""\
{d.As} are listed in a text with {d.n} and {d.p} and an optional {d.c} in {d.Pl}. \
This text can be directly edited. It can also contain blank lines and comments. \
The text can be saved, loaded and appended with with the hamburger file menu. \
The  hamburger menu has Undo and Redo buttons to return to a previous state.

It is possible to enter all {d.as_} before starting with {d.S}, or to enter \
each {d.a} when it is sold, or a mixture of both methods.

It may also be possible to make a photo of a table of prices, and to use ocr \
software to extract text from the photo, before running {d.C}. The text can \
then be loaded in {d.Pl}. Then {d.ns} and {d.ps} can be copied and pasted \
to the needed syntax.""")))

    def on_help_syntax(self, action, param):
        d = DialogWidget()
        win=self
        d.help_dialog(win, _("pricelist syntax"), f(_("""\
{d.Pl} mainly consists of a large text field, where a line can represents an \
{d.a}. Lines can also be empty or contain a comment.

A line with an {d.a} contains a {d.n} and {d.p} and an optional {d.c} \
separated by spaces. Lines starting with '#' are comments. 

The {d.p} consists of digits followed by '.' or ',' and 2 additional digits. \
Text before the {d.p} is the {d.n}. The {d.p} may be followed by digits \
that represents a {d.c} to show how often the {d.a} is currently sold.""")))

    def on_append_dialog_finish(self, dialog, task):
        file=dialog.open_finish(task)
        (iter1,iter2)=self.buffer.get_bounds()
        with open(file.get_path(), 'r') as p:
            data = p.read()
            self.buffer.insert(iter2, '\n'+data)

    def read_files(self, files):
        if files:
            with open(files[0], 'r') as p:
                data = p.read()
                self.buffer.set_text(data)

if __name__ == '__main__':
    import doctest
    from app import App, MinWindow

    doctest.testmod()

    class PricelistWindow(MinWindow):

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            app = kwargs["application"]

            sale = Sale()

            pricelist_widget = PricelistWidget(sale, win=self)
            self.set_content(pricelist_widget)

            #style_manager = app.get_style_manager()
            #style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)
            #style_manager.set_color_scheme(Adw.ColorScheme.PREFER_DARK)

            if appargs.moreargs:
                # read testdata from file
                pricelist_widget.read_files(appargs.moreargs)
            else:
                # create minimal testdata now
                article = [("Banana", 1.10, 1), ("Apple", 2.0, 0)]
                for a in article:
                    sale.main_list.append(Article( a[0], a[1], a[2] ))

    class MyApp(App):

        def on_activate(self, app):
            self.win = PricelistWindow(application=app)
            self.win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# view_switch_window.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import gi, os, sys
gi.require_version(namespace='Adw', version='1')
from gi.repository import Adw, Gtk
dir1 = os.path.dirname(os.path.realpath(__file__))
dir2 = os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.app import App
from cashbox.utils import reduce_window_size
from cashbox.read_appargs import appargs

@Gtk.Template(filename=f"{dir1}/view_switch_window.ui")
class ViewSwitchWindow(Adw.ApplicationWindow):
    __gtype_name__ = 'ViewSwitchWindow'
    stack = Gtk.Template.Child() # Adw.ViewStack

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # because MinWindow is not known in view_switch_window.blp
        # ViewSwitchWindow(MinWindow) is not possible
        # and the next 2 lines have been copied from MinWindow
        if hasattr(appargs, "t1") and appargs.t1:
            reduce_window_size(self)

if __name__ == '__main__':
    # not possible:
    #   class MyViewSwitchWindow(ViewSwitchWindow):
    # because:
    #   TypeError: Inheritance from classes with @Gtk.Template decorators is not
    #   allowed at this time

    class MyApp(App):

        def on_visible_child_name(self, stack, _):
            print("on_visible_child_name: <%s>" % (stack.get_visible_child_name()))

        def on_activate(self, app):
            win = ViewSwitchWindow(application=app)
            win.stack.connect('notify::visible-child-name', self.on_visible_child_name)
            for i in [1,2,3]:
                box=Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=12,
                            margin_end=12, margin_bottom=12,margin_start=12,spacing=12)
                win.stack.add_titled(box, "name-%s"%i, "Title %s"%i)

                l=Gtk.Label(label="Label {i}")
                box.append(l)
            win.present()

    app = MyApp()
    app.run(sys.argv)



#!/usr/bin/python3

# read_appargs.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import sys, os, pathlib, gi
from datetime import datetime
gi.require_version(namespace='Adw', version='1')
from gi.repository import GLib
dir1=os.path.dirname(os.path.realpath(__file__))
dir2=os.path.dirname(dir1)
sys.path.append(dir2)
from cashbox.utils import eprint

class AttrDict(dict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            # hasattr expects AttributeError instead of KeyError
            # (needed for doctest)
            classname = type(self).__name__
            msg = f'{repr(classname)} object has no attribute {repr(key)}'
            raise AttributeError(msg)

    def __setattr__(self, key, value):
        self[key] = value

def init_appargs():
    """
    Set defaults for global variable "appargs".
    """
    global appargs
    appargs=AttrDict()

    # set default opts
    appargs["application_id"] = "de.bschu.cashbox"
    appargs["application_version"] = "0.3"
    appargs["user_app_dir"] = os.path.join(GLib.get_user_data_dir(), appargs.application_id)
    appargs["system_app_dir"] = os.path.join("/", "usr", "share", "cashbox")
    appargs["csspath"] = os.path.join(appargs.system_app_dir, "cashbox.css")
    appargs["dark"] = None
    appargs["session"] = datetime.now().strftime('%Y-%m-%d')

    appargs["currency"] = "Dollar"
    appargs["test_small_display"] = False
    appargs["max_name_len"] = 40
    appargs["conf_suffix"] = ".cshbx"

    # overwrite with dev opts
    if dir1.endswith("/src"):
        eprint(f"WARNING: using dev opts, because running in <{dir1}>")
        appargs["csspath"] = f"{dir1}/cashbox.css"

def read_appargs(opts={}, moreargs=[]):
    global appargs

    # set command_line opts
    for i in opts:
        appargs[i] = opts[i]

    # set command_line args
    appargs["moreargs"] = moreargs

    # do opts handling
    if appargs["currency"] == "Euro":
        appargs.update({"separator":",", "symbol":"€", "digits":2, "cents":10**2})
    elif appargs["currency"] == "Dollar":
        appargs.update({"separator":".", "symbol":"$", "digits":2, "cents":10**2})
    else:
        err(f"currency option <{appargs['currency']}> not Dollar or Euro")

    pathlib.Path(appargs.user_app_dir).mkdir(parents=True, exist_ok=True)

# init global variable appargs, to be included by other modules
if "appargs" not in globals():
    init_appargs()

if __name__ == '__main__':

    # try with option --help
    # try with option --number

    import gi
    gi.require_version('Gtk', '4.0')
    gi.require_version(namespace='Adw', version='1')
    from gi.repository import Adw, Gio
    import cashbox.read_appargs

    class App(Adw.Application):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, application_id="de.bschu.cashbox",
                             flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE |
                             Gio.ApplicationFlags.HANDLES_OPEN, **kwargs)

            self.connect('activate', self.on_activate)

            self.add_main_option("number", ord("n"), GLib.OptionFlags.NONE,
                                 GLib.OptionArg.INT, "Test number", None)

            self.set_option_context_parameter_string("MOREARGS")
            self.set_option_context_summary("More Args")
            self.set_option_context_description("More Args to be used in appargs.moreargs")

        def do_command_line(self, command_line):
            opts = command_line.get_options_dict().end().unpack()
            args = command_line.get_arguments()

            read_appargs(opts, args[1:])

            self.activate()
            return 0

        def on_activate(self, app):
            for i in appargs.keys():
                print(f"appargs[{i}]=<{appargs[i]}>")
            print(f"appargs.csspath=<{appargs.csspath}>")
            assert appargs.currency == "Dollar"

    app = App()
    app.run(sys.argv)



#!/usr/bin/python3

# locale_utils.py
#
# Copyright:
#   Copyright (C) 2024 Bernd Schumacher <bernd@bschu.de>
#
# License: GPL-3.0+
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   .
#   This package is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   .
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
# Comment:
#   On Debian systems, the complete text of the GNU General
#   Public License version 3 can be found in "/usr/share/common-licenses/GPL-3".

import os, gettext, locale
from copy import copy
from inspect import currentframe

def init_locale():
    """
    Set defaults for global variable "_"
    """
    global _
    _ = gettext.gettext # allow _("...") to mark stings as translateable 

    # set locale for all categories as defined in LANG
    locale.setlocale(locale.LC_ALL, "")

    # use local po directory if exists
    locale.bindtextdomain("cashbox", "po/locale" if os.path.isdir("po/locale")
                          else "/usr/share/locale" )
    locale.textdomain("cashbox")
    gettext.bindtextdomain("cashbox", "po/locale" if os.path.isdir("po/locale")
                           else "/usr/share/locale" )
    gettext.textdomain("cashbox")

# init global variable _ to be included by other modules
if "_" not in globals():
    init_locale()

def f(s:str) -> str:
    """
    There are problems when using gettext and f-string:

        name="world"
        _(f"hello {name}")

    This is not possible, because f-string will modify the string.
    And the modified string can not be found in the translation.
    This workaround allows:

        name="world"
        f(_("hello {name}"))

    """
    frame=currentframe().f_back
    kwargs = copy(frame.f_globals)
    kwargs.update(frame.f_locals)
    return s.format(**kwargs)
